/************************************************************************
 *
 *                    CalC (Calcium Calculator)
 *                         Victor Matveev
 *
 *                            grid.cpp
 *
 *    Spatial grid definition class GridObj
 *    Boundary Condition definition array class BCarrayObj
 *    Grid stretching routines
 *    Difference scheme utility array initialization routines (initScheme)
 *
 ************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include "syntax.h"
#include "box.h"
#include "grid.h"

//*****************************************************************************
//                    G R I D   O P E R A T I O N S
//*****************************************************************************
   
 void GridObj::set_dimensions(RegionObj &reg, int nx, int ny, int nz)
 {
 nx = int(nx + 0.5);  nx |= 1;
 ny = int(ny + 0.5);  ny |= 1;
 nz = int(nz + 0.5);  nz |= 1;

 
 if (VERBOSE) { 
   fprintf(stdlog, "\n### Setting up the grid: [0..%d]", nx - 1);
   if ( DIMENSIONALITY > 1) fprintf(stdlog, " x [0..%d]", ny - 1);
   if ( DIMENSIONALITY > 2) fprintf(stdlog, " x [0..%d]", nz - 1);
   fprintf(stdlog, "\n");
 }

 xsize = nx; ysize = ny; zsize = nz;

 xysize = nx * ny;
 Size = nx * ny * nz;

 xgrid = new double[nx]; xcoord = new double[nx];
 ygrid = new double[ny]; ycoord = new double[ny];
 zgrid = new double[nz]; zcoord = new double[nz];

 dxplus = new double[nx]; dxminus = new double[nx];
 dyplus = new double[ny]; dyminus = new double[ny];
 dzplus = new double[nz]; dzminus = new double[nz];
 dxinv = new double[nx];  dyinv  = new double[ny];  dzinv   = new double[nz];
 dx2inv = new double[nx]; dy2inv  = new double[ny]; dz2inv   = new double[nz];

 double dx = 1.0; double dy = 1.0; double dz = 1.0;

 if (nx > 1) dx = (reg.get_xmax() - reg.get_xmin()) / double(nx - 1);
 if (ny > 1) dy = (reg.get_ymax() - reg.get_ymin()) / double(ny - 1);
 if (nz > 1) dz = (reg.get_zmax() - reg.get_zmin()) / double(nz - 1);

 int i;
 for (i = 0; i < nx; i++)  
    { xgrid[i] = dx; xcoord[i] = reg.get_xmin() + i * dx; }
 for (i = 0; i < ny; i++)  
    { ygrid[i] = dy; ycoord[i] = reg.get_ymin() + i * dy; }
 for (i = 0; i < nz; i++)  
    { zgrid[i] = dz; zcoord[i] = reg.get_zmin() + i * dz; }

 }

//*****************************************************************************
   
 void GridObj::set_dimensions(RegionObj &region, TokenString &params)
 {
 int  nx=0, ny=0, nz=0;
 long pos;

 int gridNum = params.token_count("grid", &pos);
 if (gridNum == 0) 
   globalError( StrCpy("Forgot to define the grid!\n    Use \"grid nx ny nz\"") );
 // else if ( gridNum > 1 ) 
 //  params.errorMessage( pos, 0, "Multiple grid definitions not allowed");

 if ( DIMENSIONALITY == 1)
   nx = params.get_int(pos + 1);
 else  if ( DIMENSIONALITY == 2)
   params.trail_pars(pos, 'i', &nx, 'i', &ny );
 else
   params.trail_pars(pos, 'i', &nx, 'i', &ny, 'i', &nz);


 if ( DIMENSIONALITY > 2) { 
   if (nz <= 0) params.errorMessage( pos + 3, 0, "Check the 3rd grid dimension"); }
 if ( DIMENSIONALITY > 1) { 
   if (ny <= 0) params.errorMessage( pos + 2, 0, "Check the 2nd grid dimension"); } 

 if (nx <= 0) params.errorMessage( pos + 1, 0, "Check the 1st grid dimension");

 set_dimensions(region, nx, ny, nz);
 }

//*****************************************************************************
   
int GridObj::location_to_index(double *coord, double *grid, int size, double x)
{
int i = 0;

while (x > coord[i] && i < (size - 1) ) i++;
if (i > 0) if (x < coord[i-1] + 0.5 * grid[i-1]) i --;

return i;
}

//*****************************************************************************

long GridObj::location_to_index(double x, double y, double z)
{
int ix = location_to_index(xcoord, xgrid, xsize, x);
int iy = location_to_index(ycoord, ygrid, ysize, y);
int iz = location_to_index(zcoord, zgrid, zsize, z);

return ix + xsize * iy + xysize * iz;
}

//*****************************************************************************

void GridObj::split(long ind, int &ix, int &iy, int &iz)
    {
    iz = ind / xysize;
    iy = (ind -= iz * xysize) / xsize;
    ix = ind - iy * xsize; 
    }

//*****************************************************************************

void GridObj::grid_stretch(double *grid, int lim, double factor, int i0, int i1)
  {
  int i;

  if (i0 > lim || i1 > lim || i0 < 0 || i1 < 0) 
    throw makeMessage("in grid_stretch(): i0=%d i1=%d N=%d --> out of bounds\n",i0,i1,lim);

  if (i1 > i0) 
    {
    if (i0 == 0) i0 ++;
    for (i = i0; i < i1; i++) grid[i] = grid[i-1] * factor;
    }

  if (i0 > i1) 
    {
    if (i0 > lim - 1) i0 = lim - 1;
    for (i = i0 - 1; i >= i1; i--) grid[i] = grid[i+1] * factor;
    }
  }

/*****************************************************************************

 Non-uniform grid in [xmin, x1] and [x2, xmax];
 Stretch intervals [0, n1] and [n2, n]; 
 dx = unstretched grid distance within [n1, n2]:

  0             n1    n2                    N
  |    |   |  | ||||||| |  |   |     |      |
  xmin          x1    x2                    xmax
  <--   X1   --><- X -><--      X2        -->

 1) X = dx (n2 - n1)

                                         m=n1-1
                                           __
 2) x1 - xmin = dx f + dx f^2 + ... = f dx \   f^m = f dx (f^n1 - 1) / (f - 1) == X1
                                           /_  
                                           m=0

                                        m=N-n2-1
                                           __
 3) xmax - x2 = dx f + dx f^2 + ... = f dx \  f^m = f dx (f^(N-n2) - 1) / (f - 1) == X2
                                           /_  
                                           m=0

                  SOLUTION:
                  ^^^^^^^^
 I.   From (3): f^(N - n2) = [ X2 (f-1) / (f dx) ] + 1

 II.  From (1): f^n1 = f^(N - X/dx) / f^(N - n2) = f^(N - X/dx) / { [ X2 (f-1) / (f dx) ] + 1 }

 III. Substitute into (2):  X1 = { f^(N - X/dx) /  { [ X2 (f-1) / (f dx) ] + 1 } - 1 } f dx / (f - 1)

 IV.  Solve III for dx, using bisection

 V.   Get n1 from (2):  n1 = log(1 + X1 (f - 1) / (dx f) ) / log f

 VI.  Get n2 from (1):  n2 = X / dx + n1

*****************************************************************************/
 
void GridObj::grid_stretch(const char *dir, double factor, double x1, double x2)
{
  int       N;
  double    *grid, *coord;
  double    eps = 1.0e-7, tiny = 1.0e-12;

  if (fabs(factor - 1) < 0.000001) return;

  if (::equal(dir, LABEL_DIM1) ) { grid = xgrid; coord = xcoord; N = xsize - 1; }
  else if (::equal(dir, LABEL_DIM2) ) { grid = ygrid; coord = ycoord; N = ysize - 1; }
  else if (::equal(dir, LABEL_DIM3) ) { grid = zgrid; coord = zcoord; N = zsize - 1; }
  else throw makeMessage("Invalid grid stretch direction %s", dir);

  if (x1 < coord[0]) x1 = coord[0];
  if (x2 < coord[0]) x2 = coord[0];
  if (x1 > coord[N]) x1 = coord[N]; 
  if (x2 > coord[N]) x2 = coord[N]; 

  if (x2 < x1) 
   throw makeMessage("3rd argument of the \"stretch\" command (%g) should be greater than the 2nd argument (%g)",x2,x1); 

  if (VERBOSE) 
    fprintf(stdlog,"  # Stretching grid with factor=%g along %s-axis leaving [%.4g,%.4g] uniform\n", 
            factor, dir, x1, x2);

  double X1 = x1 - coord[0];
  double X2 = coord[N] - x2;
  double X  = x2 - x1;
  double logf = log(factor);
  double f1f = (factor - 1) / factor;

  double dx0 = X / double(N) + tiny;                        // smallest possible dx
  double dx1 = (coord[N] - coord[0]) / double(N) - tiny;    // largest possible dx
  double f0 = X1 - dx0 / f1f * (exp( (N - X/dx0) * logf) / (1 + f1f * X2/dx0) - 1);
  double f1 = X1 - dx1 / f1f * (exp( (N - X/dx1) * logf) / (1 + f1f * X2/dx1) - 1);
  double dxn, fn;

  do     // Bisection begins
    {
    //printf("dx0=%f f0=%f   dx1=%f f1=%f\n",dx0,f0,dx1,f1);
    dxn = 0.5 * (dx0 + dx1);
    fn = X1 - dxn / f1f * (exp( (N - X/dxn) * logf ) / (1 + f1f * X2 / dxn) - 1);
    if (fn * f0 > 0) { dx0 = dxn; f0 = fn; }
      else { dx1 = dxn; f1 = fn; }
    }
  while (fabs( (dx0 - dx1) / (dx0 + dx1) ) > eps );

  int n1 = int( log(1 + X1 / dxn * f1f) / logf + 0.5 );   // step V.
  int n2 = int(X / dxn + n1 + 0.5 );                      // step VI.
  
  // if (n1 == n2) { n1--; n2++; }  ??? **** check if this is relevant

  if (n1 > 0) grid_stretch(grid, N, factor, n1, 0);
  if (n2 < N) grid_stretch(grid, N, factor, n2, N);

  int i = 0;              // RESCALE:
  double range = 0.0;

  for (i = 0; i < N; i++)  range += grid[i];
  double rescale = (coord[N] - coord[0]) / range;

  for (i = 0; i < N; i++)  grid[i] *= rescale; 
  for (i = 1; i < N; i++)  coord[i] = coord[i-1] + grid[i-1];
  grid[N] = grid[N-1];

  if (VERBOSE) {
    fprintf(stdlog,"    Uniform %s-interval = [%.4g, %.4g] (%d..%d)\n", 
          dir, coord[n1], coord[n2], n1, n2);
    fprintf(stdlog,"    Shortest dx = %.4g; dx on the boundaries: %.4g; %.4g\n", 
          grid[n1], grid[0], grid[N-1]);
  }

  return;
}

//*****************************************************************************

void GridObj::grid_stretch(TokenString &params)
{
  char    dir[MAX_TOKEN_LENGTH];  // longest is "theta"
  double  x1, x2;
  double  f = 1.05;
  int     n = params.token_count("stretch");

  if (n == 0) return;

  params.get_param("stretch.factor", &f);

  if (f < 1.0 || f > 2.0) 
    params.errorMessage( params.token_index("stretch.factor") + 2, 
                         makeMessage("Stretch factor value = %g, should be between 1.0 and 2.0", f));

  for (int i = 1; i <= n; i++)
    {
    params.trail_pars("stretch", i, 'S', dir, 'd', &x1, 'd', &x2, 'E');  
    try { grid_stretch(dir, f, x1, x2); }
    catch (char *str) { params.errorMessage( params.token_index("stretch", i) + 1, str); }
    }
}
 


//*****************************************************************************

void GridObj::initScheme() {
int i;

  if (GEOMETRY1 == 1) initSchemeCylindrical();
  else if (GEOMETRY1 == 2) initSchemeSpherical(); 
  else
    for (i = 0; i < xsize; i++) {
      dxplus[i]   = 2.0  / xgrid[i] / (xgrid[i] + xgrid[i ? i-1 : 0]);
      dxminus[i]  = 2.0 / xgrid[i ? i-1 : 0] / (xgrid[i] + xgrid[i ? i-1 : 0]);
      dxinv[i] = 1 / xgrid[i];
      dx2inv[i] = 1 / (xgrid[i] * xgrid[i]);
      }

  if (DIMENSIONALITY == 1) return;

  if (GEOMETRY2 == 3) initSchemeTheta();
  else 
    for (i = 0; i < ysize; i++)    {
      dyplus[i]   = 2.0 / ygrid[i] / (ygrid[i] + ygrid[i ? i-1 : 0]);
      dyminus[i]  = 2.0 / ygrid[i ? i-1 : 0] / (ygrid[i] + ygrid[i ? i-1 : 0]);
      dyinv[i] = 1 / ygrid[i];
      dy2inv[i] = 1 / (ygrid[i] * ygrid[i]);
      }

  if (DIMENSIONALITY == 2) return;

  for (i = 0; i < zsize; i++)    {
    dzplus[i]   = 2.0 / zgrid[i] / (zgrid[i] + zgrid[i ? i-1 : 0]);
    dzminus[i]  = 2.0 / zgrid[i ? i-1 : 0] / (zgrid[i] + zgrid[i ? i-1 : 0]);
    dzinv[i] = 1 / zgrid[i];
    dz2inv[i] = 1 / (zgrid[i] * zgrid[i]);
  }

}


//*****************************************************************************

void GridObj::initSchemeCylindrical() {
int i;

 for (i = 0; i < xsize; i++)  {
   double r = xcoord[i];
   double drplus  = xgrid[i];
   double drminus = i ? xgrid[i-1] : xgrid[i];
   double rplus  = r + 0.5 * drplus;
   double rminus = r - 0.5 * drminus;

   dxplus[i] = 2 * rplus / (r * drplus * (drplus + drminus) );
   dxminus[i] = 2 * rminus / (r * drminus * (drplus + drminus) );

   dxinv[i] = 1.0 / drplus;
   dx2inv[i] = 1.0 / (drplus * drplus);
   }

}

//*****************************************************************************

 void GridObj::initSchemeSpherical()
 { 
 int i;

 for (i = 0; i < xsize; i++)  {
   double r = xcoord[i];
   double drplus  = xgrid[i];
   double drminus = i ? xgrid[i-1] : xgrid[i];
   double rplus  = r + 0.5 * drplus;
   double rminus = r - 0.5 * drminus;
   double c = 2 / ( r * r * (drplus + drminus) );

   dxplus[i] = c * rplus * rplus / drplus;
   dxminus[i] = c * rminus * rminus / drminus;

   dxinv[i] = 1.0 / drplus;
   dx2inv[i] = 1.0 / (drplus * drplus);
   }
 }

//*****************************************************************************

 void GridObj::initSchemeTheta()
 { 
 int i;

 for (i = 0; i < ysize; i++)  {
   double theta = ycoord[i];
   double dplus  = ygrid[i];
   double dminus = i ? ygrid[i-1] : ygrid[i];
   double sinplus  = sin(theta + 0.5 * dplus);
   double sinminus = sin(theta - 0.5 * dminus);
   double c = 2 / ( sin(theta) * (dplus + dminus) );

   dyplus[i] = c * sinplus / dplus;
   dyminus[i] = c * sinminus / dminus;

   dyinv[i] = 1.0 / dplus;
   dy2inv[i] = 1.0 / (dplus * dplus);
   }
 }
//*****************************************************************************

 void GridObj::kill_grid()
 { 
 delete [] xgrid;  delete [] ygrid;   delete [] zgrid; 
 delete [] xcoord; delete [] ycoord;  delete [] zcoord;

 delete [] dxplus; delete [] dxminus; 
 delete [] dyplus; delete [] dyminus; 
 delete [] dzplus; delete [] dzminus; 

 delete [] dxinv;  delete [] dyinv;   delete [] dzinv;
 delete [] dx2inv; delete [] dy2inv;  delete [] dz2inv;
 } 

//*****************************************************************************

void GridObj::grid_clamp(TokenString &params)
{
  double  x, y, z;
  int     ix, iy, iz;
  int     n = params.token_count("clamp");

  for (int i = 1; i <= n; i++)
    {
    x = xcoord[0]; y = ycoord[0]; z = zcoord[0];
    long pos = params.token_index("clamp", i);

    if (DIMENSIONALITY == 3)  params.trail_pars(pos, 'd', &x, 'd', &y,  'd', &z,  'E');  
    else if (DIMENSIONALITY == 2)  params.trail_pars(pos, 'd', &x, 'd', &y, 'E');
    else params.trail_pars(pos, 'd', &x, 'E');

    if (  !in_bounds(x, y, z) ) {
      if (DIMENSIONALITY == 1) params.errorMessage( pos, makeMessage("Cannot clamp grid to position %g: out of bounds\n",x) );
      else if (DIMENSIONALITY == 2) params.errorMessage( pos, makeMessage("Cannot clamp grid to position (%g,%g): out of bounds\n",x,y) );
      else params.errorMessage( pos, makeMessage("Cannot clamp grid to position (%g,%g,%g): out of bounds\n",x,y,z) );
    }

    long ind = location_to_index(x, y, z);
    split(ind, ix, iy, iz);   

    if (VERBOSE) fprintf(stdlog," > clamping x-node #%d: %.5g->%.5g; intervals: {%.5g,%.5g}", 
            ix, xcoord[ix], x, (ix ? xgrid[ix-1] : 0 ), xgrid[ix] );
    xcoord[ix] = x; 
    if (ix > 1) { xcoord[ix-1] = 0.5 * (xcoord[ix-2] + xcoord[ix]);
                  xgrid[ix-2] = xcoord[ix-1] - xcoord[ix-2]; }
    if (ix) xgrid[ix-1] = xcoord[ix] - xcoord[ix-1];
    if (ix < xsize-2) { xcoord[ix+1] = 0.5 * (xcoord[ix+2] + xcoord[ix]);
                          xgrid[ix+1]  = xcoord[ix+2] - xcoord[ix+1]; }
    if (ix < xsize-1) xgrid[ix] = xcoord[ix+1] - xcoord[ix]; else xgrid[ix] = xgrid[ix-1];
    if (VERBOSE) fprintf(stdlog,"->{%.5g,%.5g}\n", (ix ? xgrid[ix-1] : 0 ), xgrid[ix] );

    
    if (VERBOSE) fprintf(stdlog," > clamping y-node #%d: %.5g->%.5g; intervals: {%.5g,%.5g}", 
            iy, ycoord[iy],  y, (iy ? ygrid[iy-1] : 0 ), ygrid[iy] );
    ycoord[iy] = y; 
    if (iy > 1) { ycoord[iy-1] = 0.5 * (ycoord[iy-2] + ycoord[iy]);
                  ygrid[iy-2] = ycoord[iy-1] - ycoord[iy-2]; }
    if (iy) ygrid[iy-1] = ycoord[iy] - ycoord[iy-1];
    if (iy < ysize-2) { ycoord[iy+1] = 0.5 * (ycoord[iy+2] + ycoord[iy]);
                          ygrid[iy+1]  = ycoord[iy+2] - ycoord[iy+1]; }
    if (iy < ysize-1) ygrid[iy] = ycoord[iy+1] - ycoord[iy]; else ygrid[iy] = ygrid[iy-1];
    if (VERBOSE) fprintf(stdlog,"->{%.5g,%.5g}\n", (iy ? ygrid[iy-1] : 0 ), ygrid[iy] );
    

    if (VERBOSE) fprintf(stdlog," > clamping z-node #%d: %.5g->%.5g; intervals: {%.5g,%.5g}", 
            iz, zcoord[iz],  z, (iz ? zgrid[iz-1] : 0 ), zgrid[iz] );
    zcoord[iz] = z; 
    if (iz > 1) { zcoord[iz-1] = 0.5 * (zcoord[iz-2] + zcoord[iz]);
                  zgrid[iz-2] = zcoord[iz-1] - zcoord[iz-2]; }
    if (iz) zgrid[iz-1] = zcoord[iz] - zcoord[iz-1];
    if (iz < zsize-2) { zcoord[iz+1] = 0.5 * (zcoord[iz+2] + zcoord[iz]);
                          zgrid[iz+1]  = zcoord[iz+2] - zcoord[iz+1]; }
    if (iz < zsize-1) zgrid[iz] = zcoord[iz+1] - zcoord[iz]; else zgrid[iz] = zgrid[iz-1];
    if (VERBOSE) fprintf(stdlog,"->{%.5g,%.5g}\n", (iz ? zgrid[iz-1] : 0 ), zgrid[iz] );
    
    }
}
//*****************************************************************************

 char GridObj::in_bounds(double x, double y, double z)
 {
   bool isOut = x < xcoord[0] || x > xcoord[xsize-1];
   if ( DIMENSIONALITY > 1) isOut = isOut || y < ycoord[0] || y > ycoord[ysize-1];
   if ( DIMENSIONALITY > 2) isOut = isOut || z < zcoord[0] || z > zcoord[zsize-1];

   return isOut ? 0: 1;
 }

//*****************************************************************************
//     B O U N D A R Y   C O N D I T I O N   D A T A   S T R U C T U R E S
//*****************************************************************************

   void  BCarrayObj::set_bc_types(TokenString &params)
   {
   bc_type_count = 0; 

   int predefined = 2;

   bc_type_num = params.token_count("bc.define") + predefined;

   bc_deriv = new double[bc_type_num]; bc_lin = new double[bc_type_num];
   bc_coef  = new double[bc_type_num]; bc_id = new char *[bc_type_num];
   bc_pump  = new double[bc_type_num];

   char s[256];
   double lin = 0, deriv = 0, coef = 0, pump = 0;

   set_bc_type(0.0, 1.0, 0.0, 0.0, "Dirichlet");
   set_bc_type(1.0, 0.0, 0.0, 0.0, "Noflux");

   for (int i = 0; i < bc_type_num - predefined; i++)
     {
     params.trail_pars("bc.define", i+1, 's', s, 'd', &deriv, 'd', &lin, 'd', &coef, 'd', &pump);
     try { set_bc_type(deriv, lin, coef, pump, s); }
     catch(char *str) { params.errorMessage( params.token_index("bc.define", i + 1) + 1, str); }
     if (VERBOSE) {
       fprintf(stdlog, "\n### Defining boundary condition %s: ",bc_id[i+predefined]);
       if (bc_deriv[i+predefined]) fprintf(stdlog, "du/dn");
       if (bc_lin[i+predefined])   fprintf(stdlog, "%+g * (u - bgr) ", bc_lin[i+predefined]);
       if (bc_pump[i+predefined])  fprintf(stdlog, "/ (1%+g*u)", bc_pump[i+predefined]);
       fprintf(stdlog, " = %g \n", bc_coef[i+predefined]);
     }
     }
   }

//*****************************************************************************

   int  BCarrayObj::bcidtoint(const char *id)
   {
   for (int i=0; i<bc_type_num; i++)
     if ( strcmp(bc_id[i], id) == 0 ) return i ;
     
   throw makeMessage("Boundary condition \"%s\" is undefined",id);
   }

//*****************************************************************************

   void  BCarrayObj::set_bc_types(int n)
   {
   bc_type_num = n;
   bc_deriv = new double[n]; bc_lin = new double[n]; bc_coef = new double[n];
   bc_pump = new double[n]; bc_id = new char *[n];
   }

//*****************************************************************************

   void  BCarrayObj::set_bc_type(double a, double b, double c, double d, const char *id)
   {
   if ( fabs(a) > 1e-12 ) 
     { b /= a; c /= a; a = 1; }  // normalization: if non-zero, derivative coefficient = 1
   else 
     if ( fabs(b) > 1e-12 ) { c /= b; b = 1; }  // if a = 0, then linear coefficient  = 1
     else
       throw makeMessage("One of the first two arguments to \"bc.define\" should be non-zero (a=%g, b=%g)", a, b);
   
   bc_deriv[bc_type_count] = a;  bc_lin[bc_type_count] = b;  
   bc_coef[bc_type_count] = c;   bc_pump[bc_type_count] = d;

   bc_id[bc_type_count] = StrCpy(id);
   bc_type_count ++;
   }

//*****************************************************************************

   void  BCarrayObj::kill_bc()
   {
   for (int i = 0; i < bc_type_num; i++) delete bc_id[i];

   delete [] bc_deriv; delete [] bc_lin; 
   delete [] bc_coef;  delete [] bc_id;

   bc_type_count = 0;  // reset the static counter in set_bc_type
   }



//*****************************************************************************
//                        S T A T U S      W I D G E T S
//*****************************************************************************
